# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimString,
    Segment,
)

logger = logging.getLogger("driverng.cn3000_6000")


class CN3000_6000SegmentH(HeaderSegment):
    sender_id = HprimMultiple(
        5,
        analyzer_name=HprimString(1),
        software_version=HprimString(2),
        analyzer_serial=HprimString(3),
        interface_version=HprimString(4),
        logical_id=HprimString(5),
        nickname=HprimString(6),
        ps_code=HprimString(7),
    )
    version = HprimString(13, default="E1394-97")

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class CN3000_6000SegmentP(Segment):
    patient_id = HprimString(5, max_length=16)
    first_name = HprimString((6, 2), max_length=20)
    last_name = HprimString((6, 3), max_length=20)

    orders = HprimForeign("O", "patient")


class CN3000_6000SegmentQ(Segment):
    starting_range_id = HprimMultiple(
        3,
        rack=HprimString(1),
        position=HprimString(2),
        sample_id=HprimString(3, max_length=15),
        sample_id_attribute=HprimString(4),
    )
    tests = HprimList(
        5, basefield=HprimMultiple(0, code=HprimString(4), name=HprimString(5))
    )
    nature_of_request_day_limit = HprimString(6, default="0")
    request_date = HprimDate(7)
    user_field_1 = HprimString(11)
    status_code = HprimString(13)


class CN3000_6000SegmentO(Segment):
    specimen_id = HprimMultiple(
        3,
        rack=HprimString(1),
        position=HprimString(2),
        sample_id=HprimString(3),
        sample_id_attribute=HprimString(4),
    )
    instrument_specimen_id = HprimMultiple(
        4,
        rack=HprimString(1),
        position=HprimString(2),
        sample_id=HprimString(3),
        sample_id_attribute=HprimString(4),
        request_flag=HprimString(5),
    )
    tests = HprimList(
        5,
        basefield=HprimMultiple(
            0, code=HprimString(4), dilution=HprimString(6), option=HprimString(7)
        ),
    )
    priority = HprimString(6)
    order_date = HprimDate(7)
    action_code = HprimString(12)

    results = HprimForeign("R", "order")


class CN3000_6000SegmentR(Segment):
    test = HprimMultiple(
        3,
        code=HprimString(4),
        parameter=HprimString(5),
        dilution=HprimString(6),
        result_type=HprimString(7),
        extended_order_request=HprimString(8),
        extended_order_result=HprimString(9),
        reflex_test_request=HprimString(10),
    )
    value = HprimString(4)
    units = HprimString(5)
    abnormal_flags = HprimMultiple(
        7,
        flag=HprimString(1),
        information=HprimString(2),
        instrument_error=HprimString(3),
    )
    completed_date = HprimDate(13)


class CN3000_6000SegmentC(Segment):
    comment_source = HprimString(3)
    information_type = HprimString((4, 1))
    reagent_comment = HprimMultiple(
        4,
        information_type=HprimString(1),
        parameter_code=HprimString(2),
        lots=HprimString(3),
    )
    control_comments = HprimList(
        4,
        basefield=HprimMultiple(
            0,
            information_type=HprimString(1),
            parameter_code=HprimString(2),
            date=HprimDate(3),
            lot=HprimString(5),
        ),
    )
    calibration_comment = HprimMultiple(
        4,
        information_type=HprimString(1),
        parameter_code=HprimString(2),
        date=HprimDate(3),
        lot=HprimString(5),
    )
    comment_type = HprimString(5)


class CN3000_6000SegmentL(Segment):
    termination_code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {"code": "cn3000_6000", "label": "Sysmex CN-3000/CN-6000 ASTM"}
    SEGMENT_CLASS_MAP = {
        "H": CN3000_6000SegmentH,
        "P": CN3000_6000SegmentP,
        "O": CN3000_6000SegmentO,
        "Q": CN3000_6000SegmentQ,
        "R": CN3000_6000SegmentR,
        "C": CN3000_6000SegmentC,
        "L": CN3000_6000SegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    sample_info_flags = [
        "Hemolytic Sample",
        "Icterus Sample",
        "Lipemic Sample",
        "Defective Sample Volume",
    ]

    def create(self, **kwargs):
        message = CN3000_6000SegmentH(type_segment="H", **kwargs)
        message.end = CN3000_6000SegmentL(type_segment="L")
        return message

    @staticmethod
    def get_specimen_id(specimen_id_return, specimen_id_inquired):
        specimen_id_return.rack = specimen_id_inquired.rack
        specimen_id_return.position = specimen_id_inquired.position
        specimen_id_return.sample_id = "{:>15.15}".format(
            specimen_id_inquired.sample_id
        )
        specimen_id_return.sample_id_attribute = (
            specimen_id_inquired.sample_id_attribute
        )

    def encode_worklist(self, worklist, context):
        # Uniquement les contextes de type query
        if not context["query"]:
            logger.debug("Only context query accepted")
            return None

        msg = self.create(date=context["now"])
        worklist = worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": worklist.IGNORE, worklist.ELSE: "N"},
            priority={"S": "S", "A": "A", worklist.ELSE: "R"},
        )
        if len(worklist) != 1:
            logger.debug("Only one query accepted")
            seg_p = msg.patients.add()
            seg_o = seg_p.orders.add()
            self.get_specimen_id(
                seg_o.specimen_id, context["query"].queries[0].starting_range_id
            )
            return msg

        seg_p = msg.patients.add(
            patient_id=worklist[0].patient_id,
            last_name=worklist[0].last_name,
            first_name=worklist[0].first_name,
        )
        for sample, action, priority in worklist[0].samples:
            seg_o = seg_p.orders.add(priority=priority, action_code=action)
            self.get_specimen_id(
                seg_o.specimen_id, context["query"].queries[0].starting_range_id
            )
            for order in sample.orders:
                t = seg_o.tests.add()
                t.code = order.code
                t.dilution = order.dilution
        return msg

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(
                    sample_id=seg_q.starting_range_id.sample_id.strip(), type="patient"
                )
            return worklist

        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                sample_id = seg_o.instrument_specimen_id.sample_id.strip()
                if seg_o.action_code == "Q":
                    control_lot = ""
                    for seg_r in seg_o.results:
                        if seg_r.commentaires:
                            for seg_c in seg_r.commentaires:
                                if seg_c.information_type == "QC":
                                    lots = []
                                    for control_comment in seg_c.control_comments:
                                        lots.append(control_comment.lot)
                                    control_lot = ", ".join(lots)

                    sample = worklist.get_or_create(
                        sample_id=sample_id, control_lot=control_lot, type="control"
                    )
                else:
                    rack = "{}-{}".format(
                        seg_o.instrument_specimen_id.rack,
                        seg_o.instrument_specimen_id.position,
                    )
                    sample = worklist.get_or_create(
                        sample_id=sample_id, rack=rack, type="patient"
                    )
                for seg_r in seg_o.results:
                    if seg_r.test.parameter in self.sample_info_flags:
                        code = seg_r.test.parameter
                        value = "YES"
                    else:
                        code = seg_r.test.code
                        value = seg_r.value.strip()
                        if value.startswith("PNG") or value.startswith("CSV"):
                            value = value.replace("\\", "/")
                    if seg_r.abnormal_flags.flag != "N":
                        flags = [seg_r.abnormal_flags.flag]
                    else:
                        flags = []
                    for information in seg_r.abnormal_flags.information.split("],["):
                        if information:
                            flags.append(information.replace("[", "").split()[0])
                    comments = []
                    for instrument_error in seg_r.abnormal_flags.instrument_error.split(
                        "],["
                    ):
                        if instrument_error:
                            instrument_error = instrument_error.replace("[", "")
                            comments.append(instrument_error.replace("]", ""))

                    result = sample.results.add(
                        code=code,
                        value=value,
                        unit=seg_r.units,
                        dilution=seg_r.test.dilution,
                        date=seg_r.completed_date,
                        flags=flags,
                        comments=comments,
                    )

                    if seg_o.action_code != "Q" and seg_r.commentaires:
                        for seg_c in seg_r.commentaires:
                            if seg_c.information_type == "LOT":
                                lots = seg_c.reagent_comment.lots.split(",")
                                for lot in lots:
                                    result.lots.add(type="reagent", lot=lot)
                            elif seg_c.information_type == "CAL":
                                result.lots.add(
                                    type="calibration",
                                    lot=seg_c.calibration_comment.lot,
                                )

        return worklist
