# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime

from driverng import MessageV2, Worklist
from driverng.exceptions import ParsingError, ProgrammingError

_TYPE_MESSAGE_QUERY = "50"
_TYPE_MESSAGE_ORDER = "50"
_TYPE_MESSAGE_RESULT = "51"
_TYPE_MESSAGE_QC = "52"


class BaseVSCubeMessage(object):
    """Décomposition du message

    data[0:][:2]    00: 2cars: Toujours début message
    data[2:][:2]    LL: 2cars; Longueur des DATAS (Max FF = 255cars)
    data[4:][:2]    01: 2cars: Toujours après LL
    data[6:][:2]    TY: 2cars: Type de message (50=ORDER,51=RESULT,52=QC)
    data[8:]        DATAS = Les données sur longueur libre (Max 255cars)"""

    TYPE_MESSAGE = None
    _NBLOCK = (
        True
    )  # Indicateur du nombre de bloc dans le message( Si TY 50 ou 51, DATAS commence par le nombre de code barre echangé)

    def get_blocks(self, data):
        """Traitement du message recu"""
        if data[:2] != "00" or data[4:][:2] != "01":
            raise ParsingError("Unexpected data")

        controle_type = data[6:][:2]
        if controle_type != self.TYPE_MESSAGE:
            raise ParsingError(
                "Expect {:s} type and {:s} found in data".format(
                    self.TYPE_MESSAGE, controle_type
                )
            )

        data_length = int(data[2:][:2], 16)
        if len(data[8:]) != data_length:
            raise ParsingError(
                "Expect {:d} length and {:d} found".format(data_length, len(data[8:]))
            )

        if self._NBLOCK:
            n_blocks = int(data[8:][:2], 16)
            return (n_blocks, data[10:])
        else:
            return (None, data[8:])

    def format_message(self, data):
        return "00{:02X}01{:s}{:s}".format(len(data), self.TYPE_MESSAGE, data)

    def to_string(self):
        raise NotImplementedError


class VSCubeTube(object):
    def __init__(self, **kwargs):
        self.code_barre = kwargs.pop("code_barre", None)
        self.inconnu = kwargs.pop("inconnu", False)


class VSCubeOrderMessage(BaseVSCubeMessage):
    TYPE_MESSAGE = _TYPE_MESSAGE_ORDER

    def __init__(self):
        self.tubes = []
        super(self.__class__, self).__init__()

    def add(self, **kwargs):
        tube = VSCubeTube(**kwargs)
        self.tubes.append(tube)
        return tube

    @staticmethod
    def parse(data):
        obj = VSCubeOrderMessage()

        nb_code_barre, blocks = obj.get_blocks(data)

        code_barre = ""
        for c in blocks:
            if c in ("\x10", "\x11"):
                obj.add(code_barre=code_barre, inconnu=True if c == "\x11" else False)
                code_barre = ""
            else:
                code_barre += c

        if code_barre != "":
            raise ParsingError("Remaining data '{:s}'".format(code_barre))
        if len(obj.tubes) != nb_code_barre:
            raise ParsingError(
                "Expect {:d} bare-code and {:d} found".format(
                    nb_code_barre, len(obj.tubes)
                )
            )
        return obj

    def to_string(self):
        data = "{:02X}".format(len(self.tubes))
        for tube in self.tubes:
            data += "{:s}{:s}".format(
                tube.code_barre, "\x11" if tube.inconnu else "\x10"
            )
        return self.format_message(data)


class VSCubeResult(object):
    FMT_DATETIME = "%d%m%y%H%M"
    FMT_DATE = "%d%m%y"
    _ARGS = {
        # key => max_length
        "code_barre": 15,
        "date": None,
        "esr": 4,
        "flags": 2,
        "rack": 4,
        "position": 2,
    }

    def __init__(self, **kwargs):

        for key in self._ARGS.keys():
            setattr(self, key, kwargs.pop(key, None))

        if kwargs:
            raise ProgrammingError("Remaining kwargs {:s}".format(kwargs))

    @staticmethod
    def parse(data):
        obj = VSCubeResult()

        obj.code_barre, buf = data.split("\x10", 1)
        obj.date = datetime.datetime.strptime(buf[:10], obj.FMT_DATETIME)
        obj.esr = buf[10:][:4]
        obj.flags = buf[14:][:2]
        obj.rack = buf[16:][:4]
        obj.position = buf[20:][:2]

        return (obj, buf[22:])

    def to_string(self):
        for key, max_length in self._ARGS.items():
            value = getattr(self, key)
            if not value:
                raise ParsingError("Field {:s} is empty".format(key))
            if max_length and len(value) > max_length:
                raise ParsingError("Field {:s} too long".format(key))

        return "{:s}\x10{:s}{:4.4s}{:2.2s}{:4.4s}{:2.2s}".format(
            self.code_barre,
            self.date.strftime(self.FMT_DATETIME),
            self.esr,
            self.flags,
            self.rack,
            self.position,
        )


class VSCubeQCResult(VSCubeResult):
    _ARGS = VSCubeResult._ARGS.copy()
    _ARGS.update(**{"lot": 6, "date_peremption": None, "valref": 4})

    @staticmethod
    def parse(data):
        obj = VSCubeQCResult()
        obj.lot = data[:6]
        obj.date_peremption = datetime.datetime.strptime(data[6:][:6], obj.FMT_DATE)
        obj.valref = data[12:][:4]

        obj.code_barre, buf = data[16:].split("\x10", 1)
        obj.date = datetime.datetime.strptime(buf[:10], obj.FMT_DATETIME)
        obj.esr = buf[10:][:4].strip()
        obj.flags = buf[14:][:2]
        obj.rack = buf[16:][:4]
        obj.position = buf[20:][:2]

        return (obj, buf[22:])

    def to_string(self):
        data = super(self.__class__, self).to_string()
        return "{:6.6s}{:s}{:4.4s}{:s}".format(
            self.lot, self.date_peremption.strftime(self.FMT_DATE), self.valref, data
        )


class VSCubeResultMessage(BaseVSCubeMessage):
    TYPE_MESSAGE = _TYPE_MESSAGE_RESULT

    def __init__(self):
        self.results = []
        super(self.__class__, self).__init__()

    def add(self, **kwargs):
        result = VSCubeResult(**kwargs)
        self.results.append(result)
        return result

    @staticmethod
    def parse(data):
        obj = VSCubeResultMessage()

        nb_result, blocks = obj.get_blocks(data)

        while blocks:
            result, blocks = VSCubeResult.parse(blocks)
            obj.results.append(result)

        if len(obj.results) != nb_result:
            raise ParsingError(
                "Expect {:d} results and {:d} found".format(nb_result, len(obj.results))
            )

        return obj

    def to_string(self):
        data = "%02X" % len(self.results)

        for resu in self.results:
            data += resu.to_string()

        return self.format_message(data)


class VSCubeQCMessage(BaseVSCubeMessage):
    TYPE_MESSAGE = _TYPE_MESSAGE_QC
    _NBLOCK = False

    def __init__(self):
        self.results = []
        super(self.__class__, self).__init__()

    def add(self, **kwargs):
        result = VSCubeQCResult(**kwargs)
        self.results.append(result)
        return result

    @staticmethod
    def parse(data):
        obj = VSCubeQCMessage()

        _, blocks = obj.get_blocks(data)

        while blocks:
            result, blocks = VSCubeQCResult.parse(blocks)
            obj.results.append(result)

        return obj

    def to_string(self):
        data = ""
        for resu in self.results:
            data += resu.to_string()
        return self.format_message(data)


class Driver(MessageV2):

    # TODO(vscube) à migrer dans param_alarme
    _COMMENT_MAP = {
        0x1: "Blood column too high",
        0x2: "Blood column too low",
        0x4: "Tube Empty",
        0x8: "Generic reading error",
        0x10: "The VES of the QC measured is within " "the acceptability range",
        0x20: "The VES of the QC measured is outside" "the acceptability range",
    }

    SPEC = {"code": "vscube", "label": "Ves-Matic Cube"}

    _TYPE_MESSAGE_MAP = {
        "50": MessageV2.QUERY,
        "51": MessageV2.RESULT,
        "52": MessageV2.QC,
    }

    _MESSAGE_TYPE_MAP = {
        MessageV2.QUERY: VSCubeOrderMessage,
        MessageV2.ORDER: VSCubeOrderMessage,
        MessageV2.RESULT: VSCubeResultMessage,
        MessageV2.QC: VSCubeQCMessage,
    }

    def create(self, message_type, **kwargs):
        try:
            cls = self._MESSAGE_TYPE_MAP[message_type]
        except KeyError:
            raise ProgrammingError("Unknown message type")
        return cls(**kwargs)

    def to_string(self, message):
        return message.to_string()

    def parse(self, data):
        lines = [li for li in data.splitlines() if li]
        if len(lines) != 1:
            raise ParsingError("Cannot parse data")
        line = lines[0]

        command = line[6:][:2]

        return self._MESSAGE_TYPE_MAP[self._TYPE_MESSAGE_MAP[command]].parse(line)

    def encode_worklist(self, worklist, context):
        msg = self.create(self.ORDER)
        requested = [s.sample_id for s in worklist.samples if s.orders]
        for sample_id in requested:
            msg.add(code_barre=sample_id)

        if requested or context["query"]:
            return msg

    def decode_message(self, message):
        if message.TYPE_MESSAGE == _TYPE_MESSAGE_QUERY:
            worklist = Worklist(context="query")
            for tube in message.tubes:
                worklist.get_or_create(sample_id=tube.code_barre, type="patient")
            return worklist
        worklist = Worklist(context="result")
        for resu in message.results:
            rack = resu.rack.strip() or None
            position = resu.position.strip()
            if rack and position:
                rack += "-{:s}".format(position)
            flags = int(resu.flags, 16)
            commentaires = []
            for key, commentaire in self._COMMENT_MAP.items():
                if flags & key:
                    commentaires.append(commentaire)

            sample = worklist.get_or_create(
                type="patient"
                if message.TYPE_MESSAGE == _TYPE_MESSAGE_RESULT
                else "control",
                sample_id=resu.code_barre,
                control_lot=resu.code_barre
                if message.TYPE_MESSAGE == _TYPE_MESSAGE_QC
                else None,
                rack=rack,
            )
            sample.results.add(
                code="VS1",
                value=resu.esr.strip(),
                date=resu.date,
                comments=commentaires,
                flags="{:X}".format(flags),
            )
        return worklist
